"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2022 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging

from . import config as cfg
from . import utilities as util
from . import dialog
from . import project_functions
from PyQt5.QtWidgets import QFileDialog


def get_info(self):
    """
    show info about media file (current media file if observation opened)
    """

    if self.observationId and self.playerType == cfg.MEDIA:

        tot_output = ""

        for i, dw in enumerate(self.dw_player):
            if not (
                str(i + 1) in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE]
                and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]
            ):
                continue

            logging.info(f"Video format: {dw.player.video_format}")
            logging.info(f"number of media in media list: {dw.player.playlist_count}")
            logging.info(f"Current time position: {dw.player.time_pos}  duration: {dw.player.duration}")

            logging.info(f"FPS: {dw.player.container_fps}")

            # logging.info("Rate: {}".format(player.mediaplayer.get_rate()))
            logging.info(f"Video size: {dw.player.width}x{dw.player.height}  ratio: ")

            logging.info(f"Aspect ratio: {round(dw.player.width / dw.player.height, 3)}")
            # logging.info("is seekable? {0}".format(player.mediaplayer.is_seekable()))
            # logging.info("has_vout? {0}".format(player.mediaplayer.has_vout()))

            mpv_output = (
                "<b>MPV information</b><br>"
                f"Video format: {dw.player.video_format}<br>"
                # "State: {}<br>"
                # "Media Resource Location: {}<br>"
                # "File name: {}<br>"
                # "Track: {}/{}<br>"
                f"Number of media in media list: {dw.player.playlist_count}<br>"
                f"Current time position: {dw.player.time_pos}<br>"
                f"duration: {dw.player.duration}<br>"
                # "Position: {} %<br>"
                f"FPS: {dw.player.container_fps}<br>"
                # "Rate: {}<br>"
                f"Video size: {dw.player.width}x{dw.player.height}<br>"
                # "Scale: {}<br>"
                f"Aspect ratio: {round(dw.player.width / dw.player.height, 3)}<br>"
                # "is seekable? {}<br>"
                # "has_vout? {}<br>"
            )

            # FFmpeg analysis
            ffmpeg_output = "<br><b>FFmpeg analysis</b><br>"

            for file_path in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]:
                media_full_path = project_functions.full_path(file_path, self.projectFileName)
                r = util.accurate_media_analysis(self.ffmpeg_bin, media_full_path)
                nframes = r["frames_number"]
                if "error" in r:
                    ffmpeg_output += "File path: {filePath}<br><br>{error}<br><br>".format(
                        filePath=media_full_path, error=r["error"]
                    )
                else:
                    ffmpeg_output += (
                        "File path: {}<br>Duration: {}<br>Bitrate: {}k<br>"
                        "FPS: {}<br>Has video: {}<br>Has audio: {}<br><br>"
                    ).format(
                        media_full_path,
                        util.convertTime(self.timeFormat, r["duration"]),
                        r["bitrate"],
                        r["fps"],
                        r["has_video"],
                        r["has_audio"],
                    )

                ffmpeg_output += f"Total duration: {util.convertTime(self.timeFormat, sum(self.dw_player[i].media_durations) / 1000)} (hh:mm:ss.sss)"

            tot_output += mpv_output + ffmpeg_output + "<br><hr>"

        self.results = dialog.Results_dialog()
        self.results.setWindowTitle(cfg.programName + " - Media file information")
        self.results.ptText.appendHtml(tot_output)
        self.results.show()

    else:  # no open observation

        fn = QFileDialog().getOpenFileName(self, "Select a media file", "", "Media files (*)")
        file_path = fn[0] if type(fn) is tuple else fn

        if file_path:
            self.results = dialog.Results_dialog()
            self.results.setWindowTitle(f"{cfg.programName} - Media file information")
            self.results.ptText.appendHtml("<br><b>FFmpeg analysis</b><hr>")
            r = util.accurate_media_analysis(self.ffmpeg_bin, file_path)
            if "error" in r:
                self.results.ptText.appendHtml(f"File path: {file_path}<br><br>{r['error']}<br><br>")
            else:
                self.results.ptText.appendHtml(
                    (
                        f"File path: {file_path}<br>"
                        f"Duration: {util.convertTime(self.timeFormat, r['duration'])}<br>"
                        f"Bitrate: {r['bitrate']}k<br>"
                        f"FPS: {r['fps']}<br>"
                        f"Has video: {r['has_video']}<br>"
                        f"Has audio: {r['has_audio']}<br><br>"
                    )
                )

            self.results.show()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2022 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

from decimal import Decimal as dec
import tablib
import logging
import os
import sys
import datetime
import pathlib
from io import StringIO
import pandas as pd

try:
    import pyreadr

    flag_pyreadr_loaded = True
except ModuleNotFoundError:
    flag_pyreadr_loaded = False

from . import dialog
from . import config as cfg
from . import utilities as util
from . import project_functions
from . import db_functions


def export_events_jwatcher(
    parameters: list, obsId: str, observation: list, ethogram: dict, file_name: str, output_format: str
) -> tuple:
    """
    export events jwatcher .dat format

    Args:
        parameters (dict): subjects, behaviors
        obsId (str): observation id
        observation (dict): observation
        ethogram (dict): ethogram of project
        file_name (str): file name for exporting events
        output_format (str): Not used for compatibility with export_events function

    Returns:
        bool: result: True if OK else False
        str: error message
    """
    for subject in parameters[cfg.SELECTED_SUBJECTS]:

        # select events for current subject
        events = []
        for event in observation[cfg.EVENTS]:
            if event[cfg.EVENT_SUBJECT_FIELD_IDX] == subject or (
                subject == cfg.NO_FOCAL_SUBJECT and event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""
            ):
                events.append(event)

        if not events:
            continue

        total_length = 0  # in seconds
        if observation[cfg.EVENTS]:
            total_length = (
                observation[cfg.EVENTS][-1][0] - observation[cfg.EVENTS][0][0]
            )  # last event time - first event time

        file_name_subject = str(pathlib.Path(file_name).parent / pathlib.Path(file_name).stem) + "_" + subject + ".dat"

        rows = ["FirstLineOfData"]  # to be completed
        rows.append("#-----------------------------------------------------------")
        rows.append(f"# Name: {pathlib.Path(file_name_subject).name}")
        rows.append("# Format: Focal Data File 1.0")
        rows.append(f"# Updated: {datetime.datetime.now().isoformat()}")
        rows.append("#-----------------------------------------------------------")
        rows.append("")
        rows.append(f"FocalMasterFile={pathlib.Path(file_name_subject).with_suffix('.fmf')}")
        rows.append("")

        rows.append(f"# Observation started: {observation['date']}")
        try:
            start_time = datetime.datetime.strptime(observation["date"], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            start_time = datetime.datetime(1970, 1, 1, 0, 0)
        start_time_epoch = int((start_time - datetime.datetime(1970, 1, 1, 0, 0)).total_seconds() * 1000)
        rows.append(f"StartTime={start_time_epoch}")

        stop_time = (start_time + datetime.timedelta(seconds=float(total_length))).isoformat()
        stop_time_epoch = int(start_time_epoch + float(total_length) * 1000)

        rows.append(f"# Observation stopped: {stop_time}")
        rows.append(f"StopTime={stop_time_epoch}")

        rows.extend([""] * 3)
        rows.append("#BEGIN DATA")
        rows[0] = f"FirstLineOfData={len(rows) + 1}"

        all_observed_behaviors = []
        mem_number_of_state_events = {}
        for event in events:
            behav_code = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

            try:
                behavior_key = [
                    ethogram[k][cfg.BEHAVIOR_KEY] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav_code
                ][0]
            except Exception:
                # coded behavior not defined in ethogram
                continue
            if [ethogram[k][cfg.TYPE] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav_code] == [
                cfg.STATE_EVENT
            ]:
                if behav_code in mem_number_of_state_events:
                    mem_number_of_state_events[behav_code] += 1
                else:
                    mem_number_of_state_events[behav_code] = 1
                # skip the STOP event in case of STATE
                if mem_number_of_state_events[behav_code] % 2 == 0:
                    continue

            rows.append(f"{int(event[cfg.EVENT_TIME_FIELD_IDX] * 1000)}, {behavior_key}")
            if (event[cfg.EVENT_BEHAVIOR_FIELD_IDX], behavior_key) not in all_observed_behaviors:
                all_observed_behaviors.append((event[cfg.EVENT_BEHAVIOR_FIELD_IDX], behavior_key))

        rows.append(f"{int(events[-1][0] * 1000)}, EOF\n")

        try:
            with open(file_name_subject, "w") as f_out:
                f_out.write("\n".join(rows))
        except Exception:
            return False, f"File DAT not created for subject {subject}: {sys.exc_info()[1]}"

        # create fmf file
        fmf_file_path = pathlib.Path(file_name_subject).with_suffix(".fmf")
        fmf_creation_answer = ""
        if fmf_file_path.exists():
            fmf_creation_answer = dialog.MessageDialog(
                cfg.programName,
                (f"The {fmf_file_path} file already exists.<br>" "What do you want to do?"),
                [cfg.OVERWRITE, "Skip file creation", cfg.CANCEL],
            )

            if fmf_creation_answer == cfg.CANCEL:
                return True, ""

        rows = []
        rows.append("#-----------------------------------------------------------")
        rows.append(f"# Name: {pathlib.Path(file_name_subject).with_suffix('.fmf').name}")
        rows.append("# Format: Focal Master File 1.0")
        rows.append(f"# Updated: {datetime.datetime.now().isoformat()}")
        rows.append("#-----------------------------------------------------------")
        for (behav, key) in all_observed_behaviors:
            rows.append(f"Behaviour.name.{key}={behav}")
            behav_description = [
                ethogram[k][cfg.DESCRIPTION] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav
            ][0]
            rows.append(f"Behaviour.description.{key}={behav_description}")

        rows.append(f"DurationMilliseconds={int(float(total_length) * 1000)}")
        rows.append("CountUp=false")
        rows.append("Question.1=")
        rows.append("Question.2=")
        rows.append("Question.3=")
        rows.append("Question.4=")
        rows.append("Question.5=")
        rows.append("Question.6=")
        rows.append("Notes=")
        rows.append("Supplementary=\n")

        if fmf_creation_answer == cfg.OVERWRITE or fmf_creation_answer == "":
            try:
                with open(fmf_file_path, "w") as f_out:
                    f_out.write("\n".join(rows))
            except Exception:
                return False, f"File FMF not created: {sys.exc_info()[1]}"

        # create FAF file
        faf_file_path = pathlib.Path(file_name_subject).with_suffix(".faf")
        faf_creation_answer = ""
        if faf_file_path.exists():
            faf_creation_answer = dialog.MessageDialog(
                cfg.programName,
                (f"The {faf_file_path} file already exists.<br>" "What do you want to do?"),
                [cfg.OVERWRITE, "Skip file creation", cfg.CANCEL],
            )
            if faf_creation_answer == cfg.CANCEL:
                return True, ""

        rows = []
        rows.append("#-----------------------------------------------------------")
        rows.append("# Name: {}".format(pathlib.Path(file_name_subject).with_suffix(".faf").name))
        rows.append("# Format: Focal Analysis Master File 1.0")
        rows.append("# Updated: {}".format(datetime.datetime.now().isoformat()))
        rows.append("#-----------------------------------------------------------")
        rows.append("FocalMasterFile={}".format(str(pathlib.Path(file_name_subject).with_suffix(".fmf"))))
        rows.append("")
        rows.append("TimeBinDuration=0.0")
        rows.append("EndWithLastCompleteBin=true")
        rows.append("")
        rows.append("ScoreFromBeginning=true")
        rows.append("ScoreFromBehavior=false")
        rows.append("ScoreFromFirstBehavior=false")
        rows.append("ScoreFromOffset=false")
        rows.append("")
        rows.append("Offset=0.0")
        rows.append("BehaviorToScoreFrom=")
        rows.append("")
        rows.append("OutOfSightCode=")
        rows.append("")
        rows.append("Report.StateNaturalInterval.Occurrence=false")
        rows.append("Report.StateNaturalInterval.TotalTime=false")
        rows.append("Report.StateNaturalInterval.Average=false")
        rows.append("Report.StateNaturalInterval.StandardDeviation=false")
        rows.append("Report.StateNaturalInterval.ProportionOfTime=false")
        rows.append("Report.StateNaturalInterval.ProportionOfTimeInSight=false")
        rows.append("Report.StateNaturalInterval.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateNaturalDuration.Occurrence=false")
        rows.append("Report.StateNaturalDuration.TotalTime=false")
        rows.append("Report.StateNaturalDuration.Average=false")
        rows.append("Report.StateNaturalDuration.StandardDeviation=false")
        rows.append("Report.StateNaturalDuration.ProportionOfTime=false")
        rows.append("Report.StateNaturalDuration.ProportionOfTimeInSight=false")
        rows.append("Report.StateNaturalDuration.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateAllInterval.Occurrence=false")
        rows.append("Report.StateAllInterval.TotalTime=false")
        rows.append("Report.StateAllInterval.Average=false")
        rows.append("Report.StateAllInterval.StandardDeviation=false")
        rows.append("Report.StateAllInterval.ProportionOfTime=false")
        rows.append("Report.StateAllInterval.ProportionOfTimeInSight=false")
        rows.append("Report.StateAllInterval.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateAllDuration.Occurrence=true")
        rows.append("Report.StateAllDuration.TotalTime=true")
        rows.append("Report.StateAllDuration.Average=true")
        rows.append("Report.StateAllDuration.StandardDeviation=false")
        rows.append("Report.StateAllDuration.ProportionOfTime=false")
        rows.append("Report.StateAllDuration.ProportionOfTimeInSight=true")
        rows.append("Report.StateAllDuration.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.EventNaturalInterval.EventCount=false")
        rows.append("Report.EventNaturalInterval.Occurrence=false")
        rows.append("Report.EventNaturalInterval.Average=false")
        rows.append("Report.EventNaturalInterval.StandardDeviation=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatEventCount=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatRate=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalOccurance=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalAverage=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalStandardDeviation=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllEventCount=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllRate=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalOccurance=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalAverage=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalStandardDeviation=false")
        rows.append("")
        rows.append("AllCodesMutuallyExclusive=true")
        rows.append("")

        for (behav, key) in all_observed_behaviors:
            rows.append(f"Behavior.isModified.{key}=false")
            rows.append(f"Behavior.isSubtracted.{key}=false")
            rows.append(f"Behavior.isIgnored.{key}=false")
            rows.append(f"Behavior.isEventAnalyzed.{key}=false")
            rows.append(f"Behavior.switchesOff.{key}=")
            rows.append("")

        if faf_creation_answer == "" or faf_creation_answer == cfg.OVERWRITE:
            try:
                with open(pathlib.Path(file_name_subject).with_suffix(".faf"), "w") as f_out:
                    f_out.write("\n".join(rows))
            except Exception:
                return False, f"File FAF not created: {sys.exc_info()[1]}"

    return True, ""


def export_tabular_events(
    parameters, obsId: str, observation: dict, ethogram: dict, file_name: str, output_format: str
) -> tuple:
    """
    export events

    Args:
        parameters (dict): subjects, behaviors
        obsId (str): observation id
        observation (dict): observation
        ethogram (dict): ethogram of project
        file_name (str): file name for exporting events
        output_format (str): output for exporting events

    Returns:
        bool: result: True if OK else False
        str: error message
    """

    total_length = f"{project_functions.observation_total_length(observation):.3f}"

    eventsWithStatus = project_functions.events_start_stop(ethogram, observation[cfg.EVENTS])

    # check max number of modifiers
    max_modifiers = 0
    for event in eventsWithStatus:
        if event[cfg.EVENT_MODIFIER_FIELD_IDX]:
            max_modifiers = max(max_modifiers, len(event[cfg.EVENT_MODIFIER_FIELD_IDX].split("|")))

    # media file number
    mediaNb = 0
    if observation[cfg.TYPE] == cfg.MEDIA:
        for player in observation[cfg.FILE]:
            mediaNb += len(observation[cfg.FILE][player])

    rows = []

    # observation id
    rows.append(["Observation id", obsId])
    rows.append([""])

    # media file name
    if observation[cfg.TYPE] == cfg.MEDIA:
        rows.append(["Media file(s)"])
    elif observation[cfg.TYPE] == cfg.LIVE:
        rows.append(["Live observation"])
    elif observation[cfg.TYPE] == cfg.IMAGES:
        rows.append(["From directories of images"])
    else:
        rows.append([""])
    rows.append([""])

    if observation[cfg.TYPE] == cfg.MEDIA:
        for player in sorted(list(observation[cfg.FILE].keys())):
            for media in observation[cfg.FILE][player]:
                rows.append([f"Player #{player}", media])

    if observation[cfg.TYPE] == cfg.IMAGES:
        for dir in observation[cfg.DIRECTORIES_LIST]:
            rows.append([f"Directory", dir])

    rows.append([""])

    # date
    rows.append(["Observation date", observation.get("date", "").replace("T", " ")])
    rows.append([""])

    # description
    rows.append(["Description", util.eol2space(observation.get("description", ""))])
    rows.append([""])

    # time offset
    rows.append(["Time offset (s)", observation.get(cfg.TIME_OFFSET, 0)])
    rows.append([""])

    # independent variables
    if cfg.INDEPENDENT_VARIABLES in observation:
        rows.extend([["independent variables"], ["variable", "value"]])

        for variable in observation[cfg.INDEPENDENT_VARIABLES]:
            rows.append([variable, observation[cfg.INDEPENDENT_VARIABLES][variable]])

    rows.append([""])

    # write table header
    col = 0
    header = ["Time"]
    if observation[cfg.TYPE] == cfg.MEDIA:
        header.extend(["Media file path", "Total length", "FPS"])
    if observation[cfg.TYPE] == cfg.IMAGES:
        header.extend(
            [
                "Image file path",
                "Image index",
            ]
        )
    if observation[cfg.TYPE] == cfg.LIVE:
        header.extend(
            [
                "Total length",
            ]
        )

    header.extend(["Subject", "Behavior", "Behavioral category"])

    behavioral_category = project_functions.behavior_category(ethogram)

    for x in range(1, max_modifiers + 1):
        header.append(f"Modifier {x}")
    header.extend(["Comment", "Status"])

    rows.append(header)

    duration1 = []  # in seconds
    if observation[cfg.TYPE] == cfg.MEDIA:
        try:
            for mediaFile in observation[cfg.FILE][cfg.PLAYER1]:
                duration1.append(observation[cfg.MEDIA_INFO][cfg.LENGTH][mediaFile])
        except KeyError:
            pass

    for event in eventsWithStatus:
        if (
            (event[cfg.EVENT_SUBJECT_FIELD_IDX] in parameters[cfg.SELECTED_SUBJECTS])
            or (event[cfg.EVENT_SUBJECT_FIELD_IDX] == "" and cfg.NO_FOCAL_SUBJECT in parameters[cfg.SELECTED_SUBJECTS])
        ) and (event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in parameters[cfg.SELECTED_BEHAVIORS]):

            fields = []
            fields.append(util.intfloatstr(str(event[cfg.EVENT_TIME_FIELD_IDX])))

            if observation[cfg.TYPE] == cfg.MEDIA:

                time_ = event[cfg.EVENT_TIME_FIELD_IDX] - observation[cfg.TIME_OFFSET]
                if time_ < 0:
                    time_ = 0

                if duration1:
                    mediaFileIdx = [idx1 for idx1, x in enumerate(duration1) if time_ >= sum(duration1[0:idx1])][-1]
                    fields.append(observation[cfg.FILE][cfg.PLAYER1][mediaFileIdx])
                    fields.append(total_length)
                    # FPS
                    try:
                        fields.append(
                            observation[cfg.MEDIA_INFO][cfg.FPS][observation[cfg.FILE][cfg.PLAYER1][mediaFileIdx]]
                        )  # fps
                    except KeyError:
                        fields.append(cfg.NA)
                else:
                    fields.append(cfg.NA)  # media file
                    fields.append(cfg.NA)  # FPS

            if observation[cfg.TYPE] == cfg.LIVE:
                fields.append(total_length)  # total length

            if observation[cfg.TYPE] == cfg.IMAGES:
                fields.append(event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_PATH]])  # image file path
                fields.append(event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_INDEX]])  # image file index

            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.SUBJECT]])
            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.BEHAVIOR_CODE]])

            # behavioral category

            try:
                behav_category = behavioral_category[event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.BEHAVIOR_CODE]]]
            except Exception:
                behav_category = ""
            fields.append(behav_category)

            # modifiers
            if max_modifiers:
                modifiers = event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.MODIFIER]].split("|")
                while len(modifiers) < max_modifiers:
                    modifiers.append("")

                for m in modifiers:
                    fields.append(m)

            # comment
            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.COMMENT]].replace(os.linesep, " "))
            # status
            fields.append(event[-1])

            rows.append(fields)

    maxLen = max([len(r) for r in rows])
    data = tablib.Dataset()

    data.title = util.safe_xl_worksheet_title(obsId, output_format)

    for row in rows:
        data.append(util.complete(row, maxLen))

    r, msg = dataset_write(data, file_name, output_format)

    return r, msg


def dataset_write(dataset, file_name, output_format):
    """
    write a tablib dataset to file in specified format

    Args:
        dataset (tablib.dataset): dataset to write
        file_name (str): file name
        output_format (str): format of output

    Returns:
        bool: result. True if OK else False
        str: error message
    """

    logging.debug("function: dataset_write")

    try:

        if output_format in ("pkl", "rds"):

            # build pandas dataframe from the tsv export of tablib dataset
            df = pd.read_csv(
                StringIO(dataset.export("tsv")),
                sep="\t",
                dtype={
                    "Observation id": str,
                    "Description": str,
                    "Start (s)": float,
                    "Stop (s)": float,
                    "Comment start": str,
                    "Comment stop": str,
                    "Image index start": float,
                    "Image index stop": float,
                    "Image file path start": str,
                    "Image file path stop": str,
                    "Behavior type": str,
                },
                parse_dates=[1],
            )

            if output_format == "pkl":
                df.to_pickle(file_name)

            if flag_pyreadr_loaded and output_format == "rds":
                pyreadr.write_rds(file_name, df)

            return True, ""

        if output_format in ("csv", "tsv", "html"):
            with open(file_name, "wb") as f:
                f.write(str.encode(dataset.export(output_format)))
            return True, ""

        if output_format in ("ods", "xls", "xlsx"):

            dataset.title = util.safe_xl_worksheet_title(dataset.title, output_format)

            with open(file_name, "wb") as f:
                f.write(dataset.export(output_format))
            return True, ""

        return False, f"Format {output_format} not found"

    except Exception:
        return False, str(sys.exc_info()[1])


def export_aggregated_events(pj: dict, parameters: dict, obsId: str):
    """
    export aggregated events of one observation

    Args:
        pj (dict): BORIS project
        parameters (dict): subjects, behaviors
        obsId (str): observation id

    Returns:
        tablib.Dataset:

    """
    logging.debug(f"function: export aggregated events {parameters} {obsId}")

    interval = parameters["time"]
    start_time = parameters[cfg.START_TIME]
    end_time = parameters[cfg.END_TIME]

    data = tablib.Dataset()
    observation = pj[cfg.OBSERVATIONS][obsId]

    # obs description
    obs_description = observation[cfg.DESCRIPTION]

    obs_length = project_functions.observation_total_length(pj[cfg.OBSERVATIONS][obsId])

    if obs_length == dec(-1):  # media length not available
        interval = cfg.TIME_EVENTS

    if obs_length == dec(-2):  # obs without timestamp
        interval = cfg.TIME_EVENTS

    logging.debug(f"obs_length: {obs_length}")

    _, _, connector = db_functions.load_aggregated_events_in_db(
        pj, parameters[cfg.SELECTED_SUBJECTS], [obsId], parameters[cfg.SELECTED_BEHAVIORS]
    )
    if connector is None:
        logging.critical(f"error when loading aggregated events in DB")
        return data

    # time
    cursor = connector.cursor()

    if interval == cfg.TIME_FULL_OBS:
        min_time = float(0)
        max_time = float(obs_length)

    if interval == cfg.TIME_EVENTS:
        try:
            min_time = float(pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][0][0])
        except Exception:
            min_time = float(0)
        try:
            max_time = float(pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0])
        except Exception:
            max_time = float(obs_length)

    if interval == cfg.TIME_ARBITRARY_INTERVAL:
        min_time = float(start_time)
        max_time = float(end_time)

    logging.debug(f"min_time: {min_time}  max_time: {max_time}")

    # adapt start and stop to the selected time interval
    if obs_length != dec(-2):
        cursor.execute(
            "UPDATE aggregated_events SET start = ? WHERE observation = ? AND start < ? AND stop BETWEEN ? AND ?",
            (
                min_time,
                obsId,
                min_time,
                min_time,
                max_time,
            ),
        )
        cursor.execute(
            "UPDATE aggregated_events SET stop = ? WHERE observation = ? AND stop > ? AND start BETWEEN ? AND ?",
            (
                max_time,
                obsId,
                max_time,
                min_time,
                max_time,
            ),
        )

        cursor.execute(
            "UPDATE aggregated_events SET start = ?, stop = ? WHERE observation = ? AND start < ? AND stop > ?",
            (
                min_time,
                max_time,
                obsId,
                min_time,
                max_time,
            ),
        )

        cursor.execute(
            "DELETE FROM aggregated_events WHERE observation = ? AND (start < ? AND stop < ?) OR (start > ? AND stop > ?)",
            (
                obsId,
                min_time,
                min_time,
                max_time,
                max_time,
            ),
        )

    behavioral_category = project_functions.behavior_category(pj[cfg.ETHOGRAM])

    for subject in parameters[cfg.SELECTED_SUBJECTS]:

        # calculate observation duration by subject (by obs)
        cursor.execute(
            ("SELECT SUM(stop - start) AS duration " "FROM aggregated_events " "WHERE subject = ? "), (subject,)
        )
        duration_by_subject_by_obs = cursor.fetchone()["duration"]
        if duration_by_subject_by_obs is not None:
            duration_by_subject_by_obs = round(duration_by_subject_by_obs, 3)

        for behavior in parameters[cfg.SELECTED_BEHAVIORS]:

            cursor.execute(
                "SELECT DISTINCT modifiers FROM aggregated_events WHERE subject=? AND behavior=? ORDER BY modifiers",
                (
                    subject,
                    behavior,
                ),
            )

            rows_distinct_modifiers = list(x[0] for x in cursor.fetchall())

            for distinct_modifiers in rows_distinct_modifiers:

                cursor.execute(
                    (
                        "SELECT start, stop, type, modifiers, comment, comment_stop, "
                        "image_index_start, image_index_stop, image_path_start, image_path_stop "
                        "FROM aggregated_events "
                        "WHERE subject = ? AND behavior = ? AND modifiers = ? ORDER BY start, image_index_start"
                    ),
                    (subject, behavior, distinct_modifiers),
                )

                for row in cursor.fetchall():

                    if observation[cfg.TYPE] == cfg.MEDIA:
                        observation_type = "Media file"
                        media_file_str = ""
                        fps_str = ""
                        for player in observation[cfg.FILE]:
                            if observation[cfg.FILE][player]:
                                media_file_str += f"player #{player}: "
                                fps_str += f"player #{player}: "
                                for media_file in observation[cfg.FILE][player]:
                                    media_file_str += f"{media_file}; "
                                    fps_str += f"{observation[cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}; "

                    if observation[cfg.TYPE] == cfg.LIVE:
                        observation_type = "Live observation"
                        media_file_str = cfg.NA
                        fps_str = cfg.NA

                    if observation[cfg.TYPE] == cfg.IMAGES:
                        observation_type = "From pictures"
                        media_file_str = ""
                        for dir in observation[cfg.DIRECTORIES_LIST]:
                            media_file_str += f"{dir}; "
                        fps_str = cfg.NA

                    row_data = []

                    row_data.extend(
                        [
                            obsId,
                            observation["date"].replace("T", " "),
                            obs_description,
                            observation_type,
                            media_file_str,
                            f"{obs_length:.3f}" if obs_length not in (dec(-1), dec(-2)) else float("NaN"),
                            fps_str,
                        ]
                    )

                    # independent variables
                    if cfg.INDEPENDENT_VARIABLES in pj:
                        for idx_var in util.sorted_keys(pj[cfg.INDEPENDENT_VARIABLES]):
                            if (
                                pj[cfg.INDEPENDENT_VARIABLES][idx_var]["label"]
                                in observation[cfg.INDEPENDENT_VARIABLES]
                            ):
                                var_value = observation[cfg.INDEPENDENT_VARIABLES][
                                    pj[cfg.INDEPENDENT_VARIABLES][idx_var]["label"]
                                ]
                                if pj[cfg.INDEPENDENT_VARIABLES][idx_var]["type"] == "timestamp":
                                    var_value = var_value.replace("T", " ")

                                row_data.append(var_value)
                            else:
                                row_data.append("")

                    if row["type"] == cfg.POINT:

                        row_data.extend(
                            [
                                subject,
                                duration_by_subject_by_obs,
                                behavior,
                                behavioral_category[behavior] if behavioral_category[behavior] else "Not defined",
                                row["modifiers"],
                                cfg.POINT,
                                f"{row['start']:.3f}" if row["start"] is not None else float("NaN"),  # start
                                f"{row['stop']:.3f}" if row["stop"] is not None else float("NaN"),  # stop
                                0,  # duration
                                row["image_index_start"],
                                row["image_index_stop"],
                                row["image_path_start"],
                                row["image_path_stop"],
                                row["comment"],  # comment start
                                "",  # comment stop
                            ]
                        )
                        data.append(row_data)

                    if row["type"] == cfg.STATE:

                        row_data.extend(
                            [
                                subject,
                                duration_by_subject_by_obs,
                                behavior,
                                behavioral_category[behavior] if behavioral_category[behavior] else "Not defined",
                                row["modifiers"],
                                cfg.STATE,
                                f"{row['start']:.3f}" if row["start"] is not None else float("NaN"),
                                f"{row['stop']:.3f}" if row["stop"] is not None else float("NaN"),
                                # duration
                                f"{row['stop'] - row['start']:.3f}"
                                if (row["stop"] is not None) and (row["start"] is not None)
                                else float("NaN"),
                                row["image_index_start"],
                                row["image_index_stop"],
                                row["image_path_start"],
                                row["image_path_stop"],
                                row["comment"],
                                row["comment_stop"],
                            ]
                        )
                        data.append(row_data)

    return data


def events_to_behavioral_sequences(pj, obs_id: str, subj: str, parameters: dict, behav_seq_separator: str) -> str:
    """
    return the behavioral sequence (behavioral string) for subject in obs_id

    Args:
        pj (dict): project
        obs_id (str): observation id
        subj (str): subject
        parameters (dict): parameters
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral string for selected subject in selected observation
    """

    out = ""
    current_states = []
    # add status (POINT, START, STOP) to event
    events_with_status = project_functions.events_start_stop(pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS])

    for event in events_with_status:
        # check if event in selected behaviors

        if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in parameters[cfg.SELECTED_BEHAVIORS]:
            continue

        if event[cfg.EVENT_SUBJECT_FIELD_IDX] == subj or (
            subj == cfg.NO_FOCAL_SUBJECT and event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""
        ):

            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.POINT:
            if event[-1] == cfg.POINT:  # status is last element
                if current_states:
                    out += "+".join(current_states) + "+" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                else:
                    out += event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    out += "&" + event[cfg.EVENT_MODIFIER_FIELD_IDX].replace("|", "+")

                out += behav_seq_separator

            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.START:
            if event[-1] == cfg.START:  # status is last element
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    current_states.append(
                        (
                            f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                            f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                            f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                        )
                    )
                else:
                    current_states.append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

                out += "+".join(sorted(current_states))

                out += behav_seq_separator

            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.STOP:
            if event[-1] == cfg.STOP:

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    behav_modif = (
                        f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                        f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                        f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                    )
                else:
                    behav_modif = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                if behav_modif in current_states:
                    current_states.remove(behav_modif)

                if current_states:
                    out += "+".join(sorted(current_states))

                    out += behav_seq_separator

    # remove last separator (if separator not empty)
    if behav_seq_separator:
        out = out[0 : -len(behav_seq_separator)]

    return out


def events_to_behavioral_sequences_all_subj(
    pj, obs_id: str, subjects_list: list, parameters: dict, behav_seq_separator: str
) -> str:
    """
    return the behavioral sequences for all selected subjects in obs_id

    Args:
        pj (dict): project
        obs_id (str): observation id
        subjects_list (list): list of subjects
        parameters (dict): parameters
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral sequences for all selected subjects in selected observation
    """

    out = ""
    current_states = {i: [] for i in subjects_list}
    events_with_status = project_functions.events_start_stop(pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS])

    for event in events_with_status:
        # check if event in selected behaviors
        if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in parameters[cfg.SELECTED_BEHAVIORS]:
            continue

        if (event[cfg.EVENT_SUBJECT_FIELD_IDX] in subjects_list) or (
            event[cfg.EVENT_SUBJECT_FIELD_IDX] == "" and cfg.NO_FOCAL_SUBJECT in subjects_list
        ):

            subject = event[cfg.EVENT_SUBJECT_FIELD_IDX] if event[cfg.EVENT_SUBJECT_FIELD_IDX] else cfg.NO_FOCAL_SUBJECT

            if event[-1] == cfg.POINT:
                if current_states[subject]:
                    out += (
                        f"[{subject}]" + "+".join(current_states[subject]) + "+" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                    )
                else:
                    out += f"[{subject}]" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    out += "&" + event[cfg.EVENT_MODIFIER_FIELD_IDX].replace("|", "+")

                out += behav_seq_separator

            if event[-1] == cfg.START:
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    current_states[subject].append(
                        (
                            f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                            f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                            f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                        )
                    )
                else:
                    current_states[subject].append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

                out += f"[{subject}]" + "+".join(sorted(current_states[subject]))

                out += behav_seq_separator

            if event[-1] == cfg.STOP:

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    behav_modif = (
                        f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                        f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                        f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                    )
                else:
                    behav_modif = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                if behav_modif in current_states[subject]:
                    current_states[subject].remove(behav_modif)

                if current_states[subject]:
                    out += f"[{subject}]" + "+".join(sorted(current_states[subject]))

                    out += behav_seq_separator

    # remove last separator (if separator not empty)
    if behav_seq_separator:
        out = out[0 : -len(behav_seq_separator)]

    return out


def events_to_timed_behavioral_sequences(
    pj: dict, obs_id: str, subject: str, parameters: dict, precision: float, behav_seq_separator: str
) -> str:
    """
    return the behavioral string for subject in obsId

    Args:
        pj (dict): project
        obs_id (str): observation id
        subj (str): subject
        parameters (dict): parameters
        precision (float): time value for scan sample
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral string for selected subject in selected observation
    """

    out = ""
    current_states = []
    # events_with_status = project_functions.events_start_stop(pj[ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS])

    state_behaviors_codes = util.state_behavior_codes(pj[cfg.ETHOGRAM])
    delta = dec(str(round(precision, 3)))
    out = ""
    t = dec("0.000")

    current = []
    while t < pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0]:
        """
        if out:
            out += behav_seq_separator
        """
        csbs = util.get_current_states_modifiers_by_subject(
            state_behaviors_codes,
            pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS],
            {"": {"name": subject}},
            t,
            include_modifiers=False,
        )[""]
        if csbs:
            if current:
                if csbs == current[-1]:
                    current.append("+".join(csbs))
                else:
                    out.append(current)
                    current = [csbs]
            else:
                current = [csbs]

        t += delta

    return out


def observation_to_behavioral_sequences(
    pj, selected_observations, parameters, behaviors_separator, separated_subjects, timed, file_name
):

    try:
        with open(file_name, "w", encoding="utf-8") as out_file:
            for obs_id in selected_observations:
                # observation id
                out_file.write("\n" + f"# observation id: {obs_id}" + "\n")
                # observation description
                descr = pj[cfg.OBSERVATIONS][obs_id]["description"]
                if "\r\n" in descr:
                    descr = descr.replace("\r\n", "\n# ")
                elif "\r" in descr:
                    descr = descr.replace("\r", "\n# ")
                out_file.write(f"# observation description: {descr}\n\n")
                # media file name
                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.MEDIA:
                    out_file.write(f"# Observation type: Media file{os.linesep}")
                    media_file_str = ""

                    for player in pj[cfg.OBSERVATIONS][obs_id][cfg.FILE]:
                        if pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][player]:
                            media_file_str += f"player #{player}: "
                            # fps_str += f"player #{player}: "
                            for media_file in pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][player]:
                                media_file_str += f"{media_file}; "
                                # fps_str += f"{pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}; "

                    out_file.write(f"# Media file path: {media_file_str}\n\n")
                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.LIVE:
                    out_file.write(f"# Observation type: Live observation{os.linesep}{os.linesep}")

                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.IMAGES:
                    out_file.write(f"# Observation type: From pictures{os.linesep}{os.linesep}")

                # independent variables
                if cfg.INDEPENDENT_VARIABLES in pj[cfg.OBSERVATIONS][obs_id]:
                    out_file.write("# Independent variables\n")

                    for variable in pj[cfg.OBSERVATIONS][obs_id][cfg.INDEPENDENT_VARIABLES]:
                        out_file.write(
                            f"# {variable}: {pj[cfg.OBSERVATIONS][obs_id][cfg.INDEPENDENT_VARIABLES][variable]}\n"
                        )
                out_file.write("\n")

                # one sequence for all subjects
                if not separated_subjects:
                    out = events_to_behavioral_sequences_all_subj(
                        pj, obs_id, parameters[cfg.SELECTED_SUBJECTS], parameters, behaviors_separator
                    )
                    if out:
                        out_file.write(out + "\n")

                # one sequence by subject
                if separated_subjects:
                    # selected subjects
                    for subject in parameters[cfg.SELECTED_SUBJECTS]:
                        out_file.write(f"\n# {subject if subject else cfg.NO_FOCAL_SUBJECT}:\n")

                        if not timed:
                            out = events_to_behavioral_sequences(pj, obs_id, subject, parameters, behaviors_separator)

                        if timed:
                            out = events_to_timed_behavioral_sequences(
                                pj, obs_id, subject, parameters, 0.001, behaviors_separator
                            )

                        if out:
                            out_file.write(out + "\n")

            return True, ""

    except Exception:
        error_type, error_file_name, error_lineno = util.error_info(sys.exc_info())
        return False, f"{error_type} {error_file_name} {error_lineno}"



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2022 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import datetime
import json
import logging
import socket

from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import QApplication, QInputDialog, QLineEdit, QListWidgetItem

from . import config as cfg
from . import dialog
from . import utilities as util


class ProjectServerThread(QThread):
    """
    thread for serving project to BORIS mobile app
    """

    signal = pyqtSignal(dict)

    def __init__(self, message):
        QThread.__init__(self)
        self.message = message

    def __del__(self):
        self.wait()

    def run(self):

        BUFFER_SIZE = 1024

        s = socket.socket()
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.settimeout(1800)

        s.bind((util.get_ip_address(), 0))
        self.signal.emit({"URL": f"{s.getsockname()[0]}:{s.getsockname()[1]}"})

        s.listen(5)
        while 1:
            try:
                c, addr = s.accept()

                logging.debug(f"Got connection from {addr}")

            except socket.timeout:
                s.close()

                logging.debug("Project server timeout")

                self.signal.emit({"MESSAGE": "Project server timeout"})
                return

            rq = c.recv(BUFFER_SIZE)

            logging.debug(f"request: {rq}")

            if rq == b"get":
                msg = self.message
                while msg:
                    c.send(msg[0:BUFFER_SIZE])
                    msg = msg[BUFFER_SIZE:]
                c.close()

                logging.debug("Project sent")

                self.signal.emit({"MESSAGE": f"Project sent to {addr[0]}"})

            if rq == b"stop":
                c.close()

                logging.debug("server stopped")

                self.signal.emit({"MESSAGE": "The server is now stopped"})
                return

            # receive an observation
            if rq == b"put":
                c.send(b"SEND")
                c.close()
                c2, addr = s.accept()
                rq2 = b""
                while 1:
                    d = c2.recv(BUFFER_SIZE)
                    if d:
                        rq2 += d
                        if rq2.endswith(b"#####"):
                            break
                    else:
                        break
                c2.close()
                self.signal.emit({"RECEIVED": f"{rq2.decode('utf-8')}", "SENDER": addr})


def send_project_via_socket(self):
    """
    send project to a device via socket
    """

    def receive_signal(msg_dict):

        if "RECEIVED" in msg_dict:
            try:
                sent_obs = json.loads(msg_dict["RECEIVED"][:-5])  # cut final
            except Exception:
                logging.debug("error receiving observation")
                del self.w
                self.actionSend_project.setText("Project server")
                return

            logging.debug(f"decoded {type(sent_obs)} length: {len(sent_obs)}")

            flag_msg = False
            mem_obsid = ""
            for obsId in sent_obs:

                self.w.lwi.addItem(
                    QListWidgetItem(f"{datetime.datetime.now().isoformat()}: Observation {obsId} received")
                )
                self.w.lwi.scrollToBottom()

                if obsId in self.pj[cfg.OBSERVATIONS]:
                    flag_msg = True
                    response = dialog.MessageDialog(
                        cfg.programName,
                        (
                            f"An observation with the same id<br><b>{obsId}</b><br>"
                            f"received from<br><b>{msg_dict['SENDER'][0]}</b><br>"
                            "already exists in the current project."
                        ),
                        [cfg.OVERWRITE, "Rename received observation", cfg.CANCEL],
                    )

                    if response == cfg.CANCEL:
                        return
                    self.project_changed()
                    if response == cfg.OVERWRITE:
                        self.pj[cfg.OBSERVATIONS][obsId] = dict(sent_obs[obsId])

                    if response == "Rename received observation":
                        new_id = obsId
                        while new_id in self.pj[cfg.OBSERVATIONS]:
                            new_id, ok = QInputDialog.getText(
                                self,
                                f"Rename observation received from {msg_dict['SENDER'][0]}",
                                "New observation id:",
                                QLineEdit.Normal,
                                new_id,
                            )

                        self.pj[cfg.OBSERVATIONS][new_id] = dict(sent_obs[obsId])

                else:
                    self.pj[cfg.OBSERVATIONS][obsId] = dict(sent_obs[obsId])
                    self.project_changed()
                    mem_obsid = obsId

        elif "URL" in msg_dict:
            self.tcp_port = int(msg_dict["URL"].split(":")[-1])
            self.w.label.setText(f"Project server URL:<br><b>{msg_dict['URL']}</b><br><br>Timeout: 30 minutes")

        else:
            if "stopped" in msg_dict["MESSAGE"] or "timeout" in msg_dict["MESSAGE"]:
                del self.w
                self.actionSend_project.setText("Project server")
            else:
                self.w.lwi.addItem(QListWidgetItem(f"{datetime.datetime.now().isoformat()}: {msg_dict['MESSAGE']}"))
                self.w.lwi.scrollToBottom()

    if "server" in self.actionSend_project.text():

        include_obs = cfg.NO
        if self.pj[cfg.OBSERVATIONS]:
            include_obs = dialog.MessageDialog(cfg.programName, "Include observations?", [cfg.YES, cfg.NO, cfg.CANCEL])
            if include_obs == cfg.CANCEL:
                return

        self.w = dialog.Info_widget()
        self.w.resize(450, 100)
        self.w.setWindowFlags(Qt.WindowStaysOnTopHint)
        self.w.setWindowTitle("Project server")
        self.w.label.setText("")
        self.w.show()
        QApplication.processEvents()

        cp_project = dict(self.pj)
        if include_obs == cfg.NO:
            cp_project[cfg.OBSERVATIONS] = {}

        self.server_thread = ProjectServerThread(
            message=str.encode(
                str(json.dumps(cp_project, indent=None, separators=(",", ":"), default=util.decimal_default))
            )
        )
        self.server_thread.signal.connect(receive_signal)

        self.server_thread.start()

        self.actionSend_project.setText("Stop serving project")

    # send stop msg to project server
    elif "serving" in self.actionSend_project.text():

        s = socket.socket()
        s.connect((util.get_ip_address(), self.tcp_port))
        s.send(str.encode("stop"))
        received = ""
        while 1:
            data = s.recv(20)  # BUFFER_SIZE = 20
            if not data:
                break
            received += data
        s.close



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'edit_event.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(413, 408)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.teTime = QtWidgets.QTimeEdit(Form)
        self.teTime.setObjectName("teTime")
        self.horizontalLayout_2.addWidget(self.teTime)
        self.dsbTime = QtWidgets.QDoubleSpinBox(Form)
        self.dsbTime.setDecimals(3)
        self.dsbTime.setMaximum(9999999.0)
        self.dsbTime.setObjectName("dsbTime")
        self.horizontalLayout_2.addWidget(self.dsbTime)
        self.pb_set_to_current_time = QtWidgets.QPushButton(Form)
        self.pb_set_to_current_time.setObjectName("pb_set_to_current_time")
        self.horizontalLayout_2.addWidget(self.pb_set_to_current_time)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_2)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.lbSubject = QtWidgets.QLabel(Form)
        self.lbSubject.setObjectName("lbSubject")
        self.horizontalLayout_4.addWidget(self.lbSubject)
        self.cobSubject = QtWidgets.QComboBox(Form)
        self.cobSubject.setObjectName("cobSubject")
        self.horizontalLayout_4.addWidget(self.cobSubject)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_5.addWidget(self.label_2)
        self.cobCode = QtWidgets.QComboBox(Form)
        self.cobCode.setObjectName("cobCode")
        self.horizontalLayout_5.addWidget(self.cobCode)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.leComment = QtWidgets.QPlainTextEdit(Form)
        self.leComment.setObjectName("leComment")
        self.verticalLayout.addWidget(self.leComment)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.pbCancel = QtWidgets.QPushButton(Form)
        self.pbCancel.setObjectName("pbCancel")
        self.horizontalLayout.addWidget(self.pbCancel)
        self.pbOK = QtWidgets.QPushButton(Form)
        self.pbOK.setDefault(True)
        self.pbOK.setObjectName("pbOK")
        self.horizontalLayout.addWidget(self.pbOK)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Edit event"))
        self.label.setText(_translate("Form", "Time"))
        self.teTime.setDisplayFormat(_translate("Form", "hh:mm:ss.zzz"))
        self.pb_set_to_current_time.setText(_translate("Form", "Set to current time"))
        self.lbSubject.setText(_translate("Form", "Subject"))
        self.label_2.setText(_translate("Form", "Code"))
        self.label_4.setText(_translate("Form", "Comment"))
        self.pbCancel.setText(_translate("Form", "Cancel"))
        self.pbOK.setText(_translate("Form", "OK"))




"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2022 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import datetime
import logging
import os
import re
import tablib
import pathlib as pl
from decimal import Decimal as dec

from . import observation_operations
from . import utilities as util
from . import config as cfg
from . import select_observations
from . import export_observation
from . import select_subj_behav
from . import project_functions
from . import dialog
from . import db_functions

from PyQt5.QtWidgets import QFileDialog, QMessageBox, QInputDialog


def export_events_as_behavioral_sequences(self, separated_subjects=False, timed=False):
    """
    export events from selected observations by subject as behavioral sequences (plain text file)
    behaviors are separated by character specified in self.behav_seq_separator (usually pipe |)
    for use with Behatrix (see http://www.boris.unito.it/pages/behatrix)

    Args:
        separated_subjects (bool):
        timed (bool):
    """

    # ask user for observations to analyze
    _, selected_observations = select_observations.select_observations(
        self.pj, cfg.MULTIPLE, "Select observations to export as behavioral sequences"
    )

    if not selected_observations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        flagShowIncludeModifiers=True,
        flagShowExcludeBehaviorsWoEvents=False,
    )

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        return

    fn = QFileDialog().getSaveFileName(
        self, "Export events as behavioral sequences", "", "Text files (*.txt);;All files (*)"
    )
    file_name = fn[0] if type(fn) is tuple else fn

    if file_name:
        r, msg = export_observation.observation_to_behavioral_sequences(
            pj=self.pj,
            selected_observations=selected_observations,
            parameters=parameters,
            behaviors_separator=self.behav_seq_separator,
            separated_subjects=separated_subjects,
            timed=timed,
            file_name=file_name,
        )
        if not r:
            logging.critical(f"Error while exporting events as behavioral sequences: {msg}")
            QMessageBox.critical(
                None,
                cfg.programName,
                f"Error while exporting events as behavioral sequences:<br>{msg}",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )


def export_tabular_events(self, mode: str = "tabular"):
    """
    export events from selected observations in various formats: TSV, CSV, ODS, XLSX, XLS, HTML

    Args:
        mode (str): export mode: must be ["tabular", "jwatcher"]
    """

    # ask user observations to analyze
    _, selectedObservations = select_observations.select_observations(
        self.pj, cfg.MULTIPLE, windows_title="Select observations for exporting events"
    )

    if not selectedObservations:
        return

    if mode == "jwatcher":
        # check if images observation in list
        max_obs_length, _ = observation_operations.observation_length(self.pj, selectedObservations)

        # exit with message if events do not have timestamp
        if max_obs_length.is_nan():
            QMessageBox.critical(
                None,
                cfg.programName,
                ("This function is not available for observations with events that do not have timestamp"),
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

    out = ""
    # check if coded behaviors are defined in ethogram
    ethogram_behavior_codes = {self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE] for idx in self.pj[cfg.ETHOGRAM]}
    behaviors_not_defined = []
    out = ""  # will contain the output
    for obs_id in selectedObservations:
        for event in self.pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]:
            if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in ethogram_behavior_codes:
                behaviors_not_defined.append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])
    if set(sorted(behaviors_not_defined)):
        out += (
            "The following behaviors are not defined in the ethogram: "
            f"<b>{', '.join(set(sorted(behaviors_not_defined)))}</b><br><br>"
        )

    # check if state events are paired
    # UNPAIRED observations must not be removed!
    not_paired_obs_list = []
    for obsId in selectedObservations:
        r, msg = project_functions.check_state_events_obs(
            obsId, self.pj[cfg.ETHOGRAM], self.pj[cfg.OBSERVATIONS][obsId], self.timeFormat
        )

        if not r:
            out += f"Observation: <strong>{obsId}</strong><br>{msg}<br>"
            not_paired_obs_list.append(obsId)

    if out:
        out = f"Some observations have UNPAIRED state events<br><br>{out}"
        self.results = dialog.Results_dialog()
        self.results.setWindowTitle(f"{cfg.programName} - Check selected observations")
        self.results.ptText.setReadOnly(True)
        self.results.ptText.appendHtml(out)
        self.results.pbSave.setVisible(False)
        self.results.pbCancel.setVisible(True)

        if not self.results.exec_():
            return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selectedObservations,
        flagShowIncludeModifiers=False,
        flagShowExcludeBehaviorsWoEvents=False,
    )

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        return

    if mode == "tabular":
        if len(selectedObservations) > 1:  # choose directory for exporting observations

            item, ok = QInputDialog.getItem(
                self,
                "Export events format",
                "Available formats",
                (
                    cfg.TSV,
                    cfg.CSV,
                    cfg.ODS,
                    cfg.XLSX,
                    cfg.XLS,
                    cfg.HTML,
                ),
                0,
                False,
            )
            if not ok:
                return
            """outputFormat = re.sub(".* \(\*\.", "", item)[:-1]"""
            outputFormat = cfg.FILE_NAME_SUFFIX[item]

            exportDir = QFileDialog().getExistingDirectory(
                self,
                "Choose a directory to export events",
                os.path.expanduser("~"),
                options=QFileDialog.ShowDirsOnly,
            )
            if not exportDir:
                return

        if len(selectedObservations) == 1:
            extended_file_formats = [
                cfg.TSV,
                cfg.CSV,
                cfg.ODS,
                cfg.XLSX,
                cfg.XLS,
                cfg.HTML,
            ]

            fileName, filter_ = QFileDialog().getSaveFileName(
                self, "Export events", "", ";;".join(extended_file_formats)
            )
            if not fileName:
                return

            # outputFormat = file_formats[extended_file_formats.index(filter_)]
            outputFormat = cfg.FILE_NAME_SUFFIX[filter_]
            if pl.Path(fileName).suffix != "." + outputFormat:
                fileName = str(pl.Path(fileName)) + "." + outputFormat
                # check if file with new extension already exists
                if pl.Path(fileName).is_file():
                    if (
                        dialog.MessageDialog(
                            cfg.programName, f"The file {fileName} already exists.", [cfg.CANCEL, cfg.OVERWRITE]
                        )
                        == cfg.CANCEL
                    ):
                        return

    if mode == "jwatcher":
        exportDir = QFileDialog().getExistingDirectory(
            self, "Choose a directory to export events", os.path.expanduser("~"), options=QFileDialog.ShowDirsOnly
        )
        if not exportDir:
            return

        outputFormat = "dat"

    mem_command = ""  # remember user choice when file already exists
    for obsId in selectedObservations:
        if len(selectedObservations) > 1 or mode == "jwatcher":
            fileName = f"{pl.Path(exportDir) / util.safeFileName(obsId)}.{outputFormat}"
            # check if file with new extension already exists
            if mem_command != cfg.OVERWRITE_ALL and pl.Path(fileName).is_file():
                if mem_command == cfg.SKIP_ALL:
                    continue
                mem_command = dialog.MessageDialog(
                    cfg.programName,
                    f"The file {fileName} already exists.",
                    [cfg.OVERWRITE, cfg.OVERWRITE_ALL, cfg.SKIP, cfg.SKIP_ALL, cfg.CANCEL],
                )
                if mem_command == cfg.CANCEL:
                    return
                if mem_command in [cfg.SKIP, cfg.SKIP_ALL]:
                    continue

        if mode == "tabular":
            export_function = export_observation.export_tabular_events
        if mode == "jwatcher":
            export_function = export_observation.export_events_jwatcher

        r, msg = export_function(
            parameters, obsId, self.pj[cfg.OBSERVATIONS][obsId], self.pj[cfg.ETHOGRAM], fileName, outputFormat
        )

        if not r and msg:
            QMessageBox.critical(None, cfg.programName, msg, QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton)


def export_aggregated_events(self):
    """
    export aggregated events for the selected observations.

    Formats can be SQL (sql), SDIS (sds) or Tabular format (tsv, csv, ods, xlsx, xls, html)
    """

    _, selectedObservations = select_observations.select_observations(
        self.pj, cfg.MULTIPLE, "Select observations for exporting events"
    )
    if not selectedObservations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selectedObservations):
        return

    # check if state events are paired
    not_ok, selectedObservations = project_functions.check_state_events(self.pj, selectedObservations)
    if not_ok or not selectedObservations:
        return

    max_obs_length, selectedObsTotalMediaLength = observation_operations.observation_length(
        self.pj, selectedObservations
    )

    logging.debug(f"max_obs_length:{max_obs_length}  selectedObsTotalMediaLength:{selectedObsTotalMediaLength}")

    if max_obs_length == dec(-1):  # media length not available, user choose to not use events
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The duration of one or more observation is not available"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selectedObservations,
        maxTime=max_obs_length if len(selectedObservations) > 1 else selectedObsTotalMediaLength,
        flagShowIncludeModifiers=False,
        flagShowExcludeBehaviorsWoEvents=False,
    )

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        return

    # check for grouping results
    flag_group = True
    if len(selectedObservations) > 1:
        flag_group = (
            dialog.MessageDialog(
                cfg.programName, "Group events from selected observations in one file?", [cfg.YES, cfg.NO]
            )
            == cfg.YES
        )

    extended_file_formats = [
        "Tab Separated Values (*.tsv)",
        "Comma Separated Values (*.csv)",
        "Open Document Spreadsheet ODS (*.ods)",
        "Microsoft Excel Spreadsheet XLSX (*.xlsx)",
        "Legacy Microsoft Excel Spreadsheet XLS (*.xls)",
        "HTML (*.html)",
        "SDIS (*.sds)",
        "SQL dump file (*.sql)",
        "Pandas DataFrame (*.pkl)",
        "R dataframe (*.rds)",
    ]

    if flag_group:
        file_formats = [
            "tsv",
            "csv",
            "ods",
            "xlsx",
            "xls",
            "html",
            "sds",
            "sql",
            "pkl",
            "rds",
        ]  # must be in same order than extended_file_formats

        fileName, filter_ = QFileDialog().getSaveFileName(
            self, "Export aggregated events", "", ";;".join(extended_file_formats)
        )

        if not fileName:
            return

        outputFormat = file_formats[extended_file_formats.index(filter_)]
        if pl.Path(fileName).suffix != "." + outputFormat:
            # check if file with new extension already exists
            fileName = str(pl.Path(fileName)) + "." + outputFormat
            if pl.Path(fileName).is_file():
                if (
                    dialog.MessageDialog(
                        cfg.programName, f"The file {fileName} already exists.", [cfg.CANCEL, cfg.OVERWRITE]
                    )
                    == cfg.CANCEL
                ):
                    return

    else:  # not grouping

        items = (
            "Tab Separated Values (*.tsv)",
            "Comma Separated values (*.csv)",
            "Open Document Spreadsheet (*.ods)",
            "Microsoft Excel Spreadsheet XLSX (*.xlsx)",
            "Legacy Microsoft Excel Spreadsheet XLS (*.xls)",
            "HTML (*.html)",
            "SDIS (*.sds)",
            "Timed Behavioral Sequences (*.tbs)",
            "Pandas DataFrame (*.pkl)",
            "R dataframe (*.rds)",
        )
        item, ok = QInputDialog.getItem(self, "Export events format", "Available formats", items, 0, False)
        if not ok:
            return
        # read the output format code
        outputFormat = re.sub(".* \(\*\.", "", item)[:-1]

        exportDir = QFileDialog().getExistingDirectory(
            self, "Choose a directory to export events", os.path.expanduser("~"), options=QFileDialog.ShowDirsOnly
        )
        if not exportDir:
            return

    if outputFormat == "sql":
        _, _, conn = db_functions.load_aggregated_events_in_db(
            self.pj, parameters[cfg.SELECTED_SUBJECTS], selectedObservations, parameters[cfg.SELECTED_BEHAVIORS]
        )
        try:
            with open(fileName, "w") as f:
                for line in conn.iterdump():
                    f.write(f"{line}\n")
        except Exception:

            QMessageBox.critical(
                None,
                cfg.programName,
                f"The file {fileName} can not be saved",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )

        return

    header = [
        "Observation id",
        "Observation date",
        "Description",
        "Observation type",
        "Source",
        "Total length",
        "FPS",
    ]
    if cfg.INDEPENDENT_VARIABLES in self.pj:
        for idx in util.sorted_keys(self.pj[cfg.INDEPENDENT_VARIABLES]):
            header.append(self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"])

    header.extend(
        [
            "Subject",
            "Observation duration by subject by observation",
            "Behavior",
            "Behavioral category",
            "Modifiers",
            "Behavior type",
            "Start (s)",
            "Stop (s)",
            "Duration (s)",
            "Image index start",  # add image index and image file path to header
            "Image index stop",
            "Image file path start",
            "Image file path stop",
            "Comment start",
            "Comment stop",
        ]
    )

    data = tablib.Dataset()
    # sort by start time
    start_idx = -9
    stop_idx = -8
    obs_id_idx = 0

    # TODO: check indep var type for dataframe output
    """ to be finished
    column_type = {"Observation id": "str", "Description": "str", "Comment start": "str", "Comment stop": "str"}
    for idx_var in util.sorted_keys(pj[cfg.INDEPENDENT_VARIABLES]):
        column_type[self.pj[cfg.INDEPENDENT_VARIABLES][idx_var]["label"]] = 
        if self.pj[cfg.INDEPENDENT_VARIABLES][idx_var]["type"] == "timestamp":
    """

    mem_command = ""  # remember user choice when file already exists
    for obs_id in selectedObservations:
        d = export_observation.export_aggregated_events(self.pj, parameters, obs_id)
        data.extend(d)

        if not flag_group and outputFormat not in ["sds", "tbs"]:
            fileName = f"{pl.Path(exportDir) / util.safeFileName(obs_id)}.{outputFormat}"
            # check if file with new extension already exists
            if mem_command != cfg.OVERWRITE_ALL and pl.Path(fileName).is_file():
                if mem_command == "Skip all":
                    continue
                mem_command = dialog.MessageDialog(
                    cfg.programName,
                    f"The file {fileName} already exists.",
                    [cfg.OVERWRITE, cfg.OVERWRITE_ALL, "Skip", "Skip all", cfg.CANCEL],
                )
                if mem_command == cfg.CANCEL:
                    return
                if mem_command in ["Skip", "Skip all"]:
                    continue

            data = tablib.Dataset(
                *sorted(list(data), key=lambda x: (x[obs_id_idx], float(x[start_idx]))), headers=header
            )
            data.title = obs_id
            r, msg = export_observation.dataset_write(data, fileName, outputFormat)
            if not r:
                QMessageBox.warning(
                    None, cfg.programName, msg, QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton
                )
            data = tablib.Dataset()

    data = tablib.Dataset(*sorted(list(data), key=lambda x: (x[obs_id_idx], float(x[start_idx]))), headers=header)
    data.title = "Aggregated events"

    # TODO: finish
    if outputFormat == "tbs":  # Timed behavioral sequences
        out = ""
        for obsId in selectedObservations:
            # observation id
            out += f"# {obsId}\n"

            for event in list(data):
                if event[0] == obsId:
                    behavior = event[-8]
                    # replace various char by _
                    for char in [" ", "-", "/"]:
                        behavior = behavior.replace(char, "_")
                    subject = event[-9]
                    # replace various char by _
                    for char in [" ", "-", "/"]:
                        subject = subject.replace(char, "_")
                    event_start = f"{float(event[start_idx]):.3f}"  # start event (from end for independent variables)
                    if not event[stop_idx]:  # stop event (from end)
                        event_stop = f"{float(event[start_idx]) + 0.001:.3f}"
                    else:
                        event_stop = f"{float(event[stop_idx]):.3f}"

                    bs_timed = [f"{subject}_{behavior}"] * round((float(event_stop) - float(event_start)) * 100)
                    out += "|".join(bs_timed)

            out += "\n"

            if not flag_group:
                fileName = f"{pl.Path(exportDir) / util.safeFileName(obsId)}.{outputFormat}"
                with open(fileName, "wb") as f:
                    f.write(str.encode(out))
                out = ""

        if flag_group:
            with open(fileName, "wb") as f:
                f.write(str.encode(out))
        return

    if outputFormat == "sds":  # SDIS format
        out = ("% SDIS file created by BORIS (www.boris.unito.it) " "at {}\nTimed <seconds>;\n").format(
            util.datetime_iso8601(datetime.datetime.now())
        )
        for obsId in selectedObservations:
            # observation id
            out += "\n<{}>\n".format(obsId)

            for event in list(data):
                if event[0] == obsId:
                    behavior = event[-8]
                    # replace various char by _
                    for char in [" ", "-", "/"]:
                        behavior = behavior.replace(char, "_")
                    subject = event[-9]
                    # replace various char by _
                    for char in [" ", "-", "/"]:
                        subject = subject.replace(char, "_")
                    event_start = "{0:.3f}".format(
                        float(event[start_idx])
                    )  # start event (from end for independent variables)
                    if not event[stop_idx]:  # stop event (from end)
                        event_stop = "{0:.3f}".format(float(event[start_idx]) + 0.001)
                    else:
                        event_stop = "{0:.3f}".format(float(event[stop_idx]))
                    out += f"{subject}_{behavior},{event_start}-{event_stop} "

            out += "/\n\n"
            if not flag_group:
                """
                fileName = str(pathlib.Path(pathlib.Path(exportDir) / safeFileName(obsId)).with suffix("." + outputFormat))
                """
                fileName = f"{pl.Path(exportDir) / util.safeFileName(obsId)}.{outputFormat}"
                with open(fileName, "wb") as f:
                    f.write(str.encode(out))
                out = (
                    "% SDIS file created by BORIS (www.boris.unito.it) "
                    f"at {util.datetime_iso8601(datetime.datetime.now())}\nTimed <seconds>;\n"
                )

        if flag_group:
            with open(fileName, "wb") as f:
                f.write(str.encode(out))
        return

    if flag_group:
        r, msg = export_observation.dataset_write(data, fileName, outputFormat)
        if not r:
            QMessageBox.warning(None, cfg.programName, msg, QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton)


def export_events_as_textgrid(self):
    """
    export state events as Praat textgrid
    """

    _, selected_observations = select_observations.select_observations(self.pj, mode=cfg.MULTIPLE, windows_title="")

    if not selected_observations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    max_obs_length, _ = observation_operations.observation_length(self.pj, selected_observations)

    # exit with message if events do not have timestamp
    if max_obs_length.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    plot_parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        flagShowIncludeModifiers=False,
        flagShowExcludeBehaviorsWoEvents=False,
    )

    if not plot_parameters[cfg.SELECTED_SUBJECTS] or not plot_parameters[cfg.SELECTED_BEHAVIORS]:
        return

    exportDir = QFileDialog(self).getExistingDirectory(
        self, "Export events as Praat TextGrid", os.path.expanduser("~"), options=QFileDialog(self).ShowDirsOnly
    )
    if not exportDir:
        return

    mem_command = ""

    # see https://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html

    interval_subject_header = (
        "    item [{subject_index}]:\n"
        '        class = "IntervalTier"\n'
        '        name = "{subject}"\n'
        "        xmin = {intervalsMin}\n"
        "        xmax = {intervalsMax}\n"
        "        intervals: size = {intervalsSize}\n"
    )

    interval_template = (
        "        intervals [{count}]:\n"
        "            xmin = {xmin}\n"
        "            xmax = {xmax}\n"
        '            text = "{name}"\n'
    )

    point_subject_header = (
        "    item [{subject_index}]:\n"
        '        class = "TextTier"\n'
        '        name = "{subject}"\n'
        "        xmin = {intervalsMin}\n"
        "        xmax = {intervalsMax}\n"
        "        points: size = {intervalsSize}\n"
    )

    point_template = "        points [{count}]:\n" "            number = {number}\n" '            mark = "{mark}"\n'

    for obs_id in selected_observations:

        flagUnpairedEventFound = False

        total_media_duration = round(project_functions.observation_total_length(self.pj[cfg.OBSERVATIONS][obs_id]), 3)

        cursor = db_functions.load_events_in_db(
            self.pj,
            plot_parameters[cfg.SELECTED_SUBJECTS],
            selected_observations,
            plot_parameters[cfg.SELECTED_BEHAVIORS],
            time_interval=cfg.TIME_FULL_OBS,
        )

        cursor.execute(
            (
                "SELECT COUNT(DISTINCT subject) FROM events "
                "WHERE observation = ? AND subject IN ({}) AND type = 'STATE' ".format(
                    ",".join(["?"] * len(plot_parameters[cfg.SELECTED_SUBJECTS]))
                )
            ),
            [obs_id] + plot_parameters[cfg.SELECTED_SUBJECTS],
        )

        subjectsNum = int(list(cursor.fetchall())[0][0])

        subjectsMin, subjectsMax = 0, total_media_duration

        out = (
            'File type = "ooTextFile"\n'
            'Object class = "TextGrid"\n'
            "\n"
            f"xmin = {subjectsMin}\n"
            f"xmax = {subjectsMax}\n"
            "tiers? <exists>\n"
            f"size = {subjectsNum}\n"
            "item []:\n"
        )

        subject_index = 0
        for subject in plot_parameters[cfg.SELECTED_SUBJECTS]:
            if subject not in [x[cfg.EVENT_SUBJECT_FIELD_IDX] for x in self.pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]]:
                continue

            intervalsMin, intervalsMax = 0, total_media_duration

            # STATE events
            cursor.execute(
                (
                    "SELECT occurence, code FROM events "
                    "WHERE observation = ? AND subject = ? AND type = 'STATE' ORDER BY occurence"
                ),
                (obs_id, subject),
            )

            rows = [{"occurence": util.float2decimal(r["occurence"]), "code": r["code"]} for r in cursor.fetchall()]
            if not rows:
                continue

            out += interval_subject_header

            count = 0

            # check if 1st behavior starts at the beginning

            if rows[0]["occurence"] > 0:
                count += 1
                out += interval_template.format(count=count, name="null", xmin=0.0, xmax=rows[0]["occurence"])

            for idx, row in enumerate(rows):
                if idx % 2 == 0:

                    # check if events not interlacced
                    if row["code"] != rows[idx + 1]["code"]:
                        QMessageBox.critical(
                            None,
                            cfg.programName,
                            "The events are interlaced. It is not possible to produce the Praat TextGrid file",
                            QMessageBox.Ok | QMessageBox.Default,
                            QMessageBox.NoButton,
                        )
                        return

                    count += 1
                    out += interval_template.format(
                        count=count, name=row["code"], xmin=row["occurence"], xmax=rows[idx + 1]["occurence"]
                    )

                    # check if difference is > 0.001
                    if len(rows) > idx + 2:
                        if rows[idx + 2]["occurence"] - rows[idx + 1]["occurence"] > 0.001:

                            out += interval_template.format(
                                count=count + 1,
                                name="null",
                                xmin=rows[idx + 1]["occurence"],
                                xmax=rows[idx + 2]["occurence"],
                            )
                            count += 1
                        else:
                            rows[idx + 2]["occurence"] = rows[idx + 1]["occurence"]

            # check if last event ends at the end of media file
            if rows[-1]["occurence"] < project_functions.observation_total_length(self.pj[cfg.OBSERVATIONS][obs_id]):
                count += 1
                out += interval_template.format(
                    count=count, name="null", xmin=rows[-1]["occurence"], xmax=total_media_duration
                )

            # add info
            subject_index += 1
            out = out.format(
                subject_index=subject_index,
                subject=subject,
                intervalsSize=count,
                intervalsMin=intervalsMin,
                intervalsMax=intervalsMax,
            )

            # POINT events
            cursor.execute(
                (
                    "SELECT occurence, code FROM events "
                    "WHERE observation = ? AND subject = ? AND type = 'POINT' ORDER BY occurence"
                ),
                (obs_id, subject),
            )

            rows = [{"occurence": util.float2decimal(r["occurence"]), "code": r["code"]} for r in cursor.fetchall()]
            if not rows:
                continue

            out += point_subject_header

            count = 0

            for idx, row in enumerate(rows):

                count += 1
                out += point_template.format(count=count, mark=row["code"], number=row["occurence"])

            # add info
            subject_index += 1
            out = out.format(
                subject_index=subject_index,
                subject=subject,
                intervalsSize=count,
                intervalsMin=intervalsMin,
                intervalsMax=intervalsMax,
            )

        # check if file already exists
        if (
            mem_command != cfg.OVERWRITE_ALL
            and pl.Path(f"{pl.Path(exportDir) / util.safeFileName(obs_id)}.textGrid").is_file()
        ):
            if mem_command == "Skip all":
                continue
            mem_command = dialog.MessageDialog(
                cfg.programName,
                f"The file <b>{pl.Path(exportDir) / util.safeFileName(obs_id)}.textGrid</b> already exists.",
                [cfg.OVERWRITE, cfg.OVERWRITE_ALL, "Skip", "Skip all", cfg.CANCEL],
            )
            if mem_command == cfg.CANCEL:
                return
            if mem_command in ["Skip", "Skip all"]:
                continue

        try:
            with open(f"{pl.Path(exportDir) / util.safeFileName(obs_id)}.textGrid", "w") as f:
                f.write(out)

            if flagUnpairedEventFound:
                QMessageBox.warning(
                    self,
                    cfg.programName,
                    "Some state events are not paired. They were excluded from export",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )

        except Exception:
            QMessageBox.critical(self, cfg.programName, "The file can not be saved")
